@@ -2,7 +2,7 @@
 
     mails.c
 
-    Copyright (c) 1998-2001 Minero Aoki <aamine@loveruby.net>
+    Copyright (c) 1998-2002 Minero Aoki <aamine@loveruby.net>
 
     This program is free software.
     You can distribute/modify this program under the terms of
@@ -16,15 +16,15 @@
 #endif
 
 #include "ruby.h"
+#include "re.h"
 
 
-#define TMAIL_VERSION "0.9.10"
+#define TMAIL_VERSION "0.10.1"
 
 static VALUE MailScanner;
 static VALUE ScanError;
 
-
-struct mails
+struct scanner
 {
     char *pbeg;
     char *p;
@@ -33,31 +33,21 @@ struct mails
     VALUE comments;
 };
 
-#define MODE_MIME (1 << 0)
-#define MODE_RECV (1 << 1)
+#define MODE_MIME     (1 << 0)
+#define MODE_RECV     (1 << 1)
+#define MODE_ISO2022  (1 << 2)
+#define MODE_DEBUG    (1 << 4)
 
-#define MIME_MODE_P(s) ((s)->flags & MODE_MIME)
-#define RECV_MODE_P(s) ((s)->flags & MODE_RECV)
+#define MIME_MODE_P(s)     ((s)->flags & MODE_MIME)
+#define RECV_MODE_P(s)     ((s)->flags & MODE_RECV)
+#define ISO2022_MODE_P(s)  ((s)->flags & MODE_ISO2022)
 
-#define MAILS_DEBUG (1 << 4)
+#define GET_SCANNER(val, s) Data_Get_Struct(val, struct scanner, s)
 
-#define GET_SCANNER(val, s) Data_Get_Struct(val, struct mails, s)
-
-
-static void
-real_reset(sc)
-    struct mails *sc;
-{
-    sc->pbeg = 0;
-    sc->p = 0;
-    sc->pend = 0;
-    sc->flags = 0;
-    sc->comments = Qnil;
-}
 
 static void
 mails_free(sc)
-    struct mails *sc;
+    struct scanner *sc;
 {
     free(sc);
 }
@@ -66,29 +56,34 @@ static VALUE
 mails_s_new(klass, str, ident, cmt)
     VALUE klass;
 {
-    struct mails *sc;
-    char *tmp;
+    struct scanner *sc;
+    const char *tmp;
 
-    sc = ALLOC_N(struct mails, 1);
-    real_reset(sc);
+    sc = ALLOC_N(struct scanner, 1);
 
     Check_Type(str, T_STRING);
-    sc->pbeg = sc->p = RSTRING(str)->ptr;
+    sc->pbeg = RSTRING(str)->ptr;
+    sc->p    = sc->pbeg;
     sc->pend = sc->p + RSTRING(str)->len;
 
-    tmp = STR2CSTR(ident);
-    if      (strcmp(tmp, "RecvH") == 0)         sc->flags |= MODE_RECV;
-    else if (strcmp(tmp, "CTypeH") == 0)        sc->flags |= MODE_MIME;
-    else if (strcmp(tmp, "CEncodingH") == 0)    sc->flags |= MODE_MIME;
-    else if (strcmp(tmp, "CDispositionH") == 0) sc->flags |= MODE_MIME;
+    sc->flags = 0;
+    Check_Type(ident, T_SYMBOL);
+    tmp = rb_id2name(SYM2ID(ident));
+    if      (strcmp(tmp, "RECEIVED")     == 0) sc->flags |= MODE_RECV;
+    else if (strcmp(tmp, "CTYPE")        == 0) sc->flags |= MODE_MIME;
+    else if (strcmp(tmp, "CENCODING")    == 0) sc->flags |= MODE_MIME;
+    else if (strcmp(tmp, "CDISPOSITION") == 0) sc->flags |= MODE_MIME;
 
+    tmp = rb_get_kcode();
+    if (strcmp(tmp, "EUC") == 0 || strcmp(tmp, "SJIS") == 0) {
+        sc->flags |= MODE_ISO2022;
+    }
+
+    sc->comments = Qnil;
     if (! NIL_P(cmt)) {
         Check_Type(cmt, T_ARRAY);
         sc->comments = cmt;
     }
-    else {
-        sc->comments = Qnil;
-    }
 
     return Data_Wrap_Struct(MailScanner, 0, mails_free, sc);
 }
@@ -97,10 +92,10 @@ static VALUE
 mails_debug_get(self)
     VALUE self;
 {
-    struct mails *sc;
+    struct scanner *sc;
 
     GET_SCANNER(self, sc);
-    if (sc->flags & MAILS_DEBUG)
+    if (sc->flags & MODE_DEBUG)
         return Qtrue;
     else
         return Qfalse;
@@ -110,223 +105,180 @@ static VALUE
 mails_debug_set(self, flag)
     VALUE self, flag;
 {
-    struct mails *sc;
+    struct scanner *sc;
 
     GET_SCANNER(self, sc);
     if (RTEST(flag))
-        sc->flags |= MAILS_DEBUG;
+        sc->flags |= MODE_DEBUG;
     else
-        sc->flags &= ~MAILS_DEBUG;
+        sc->flags &= ~MODE_DEBUG;
     return Qnil;
 }
 
 
 /*
---------------------------------------------------------
-                      scan functions
---------------------------------------------------------
+----------------------------------------------------------------------
+                        scanning routines
+----------------------------------------------------------------------
 */
 
+#define ESC               '\033'
+#define ATOM_SYMBOLS      "_#!$%&'`*+-{|}~^/=?"
+#define TOKEN_SYMBOLS     "_#!$%&'`*+-{|}~^."
+#define ATOM_SPECIAL      "()<>[]@,;:\"\\."
+#define TOKEN_SPECIAL     "()<>[]@,;:\"\\/?="
+#define LWSP              " \t\r\n"
+
+#define IS_ALPHA(ch)      (('a' <= ch && ch <= 'z') || ('A' <= ch && ch <= 'Z'))
+#define IS_UPPER(ch)      ('A' <= ch && ch <= 'Z')
+#define TO_LOWER(ch)      (IS_UPPER(ch) ? ch + 32 : ch)
+#define IS_LWSP(ch)       (strchr(LWSP, ch))
+#define IS_DIGIT(ch)      ('0' <= ch && ch <= '9')
+#define IS_WORDCHAR(ch, symlist) \
+                          (IS_ALPHA(ch) || IS_DIGIT(ch) || strchr(symlist, ch))
+#define IS_ATOMCHAR(ch)   IS_WORDCHAR(ch, ATOM_SYMBOLS)
+#define IS_TOKENCHAR(ch)  IS_WORDCHAR(ch, TOKEN_SYMBOLS)
+#define IS_JCHAR(ch)      ismbchar(ch)
+
+
 /* I know this implement is agree, but usually useful. */
 
-#define ESC '\033'
-
-/* skip until "\e(B" (ascii) */
+/* skip until "\e(B" (us-ascii) */
 static void
-fwd_jis(sc)
-    struct mails *sc;
+skip_iso2022jp_string(sc)
+    struct scanner *sc;
 {
     for (; sc->p < sc->pend; sc->p++) {
-        if ((*sc->p == ESC) && (strcmp(sc->p, "\033(B") == 0)) {
-            sc->p += 3;
-            break;
+        if (*sc->p == ESC) {
+            if (strncmp(sc->p, "\033(B", 3) == 0) {
+                sc->p += 3;
+                return;
+            }
         }
     }
 }
 
-#define IS_JCHAR(ch) ((unsigned char)(ch) > 127)
-
 static void
-fwd_jstr(sc)
-    struct mails *sc;
+skip_japanese_string(sc)
+    struct scanner *sc;
 {
     while (sc->p < sc->pend) {
-        if ((unsigned char)*sc->p > 127) {
-            sc->p++;
-            if (sc->p < sc->pend) sc->p++;
-        }
-        else return;
+        if (! ismbchar(*sc->p)) return;
+        sc->p += mbclen(*sc->p);
     }
 }
 
 
+#define scan_atom(sc) scan_word(sc, ATOM_SYMBOLS)
+#define scan_token(sc) scan_word(sc, TOKEN_SYMBOLS)
+
 static VALUE
-scan_word(sc, special)
-    struct mails *sc;
-    char *special;
+scan_word(sc, syms)
+    struct scanner *sc;
+    char *syms;
 {
     char *beg = sc->p;
 
     while (sc->p < sc->pend) {
-        unsigned char ch = (unsigned char)(*sc->p);
-
-        if (strchr(special, ch)) {
+        if (ISO2022_MODE_P(sc) && *sc->p == ESC) {
+            skip_iso2022jp_string(sc);
+        }
+        else if (IS_JCHAR(*sc->p)) {
+            skip_japanese_string(sc);
+        }
+        else if (IS_WORDCHAR(*sc->p, syms)) {
+            sc->p++;
+        }
+        else {
             break;
         }
-        else if (ch == ESC) {
-            fwd_jis(sc);
-            continue;
-        }
-        else if (ch > 127) {
-            fwd_jstr(sc);
-            continue;
-        }
-        else if (ch <= 32) {
-            break;
-        }
-        sc->p++;
     }
 
     return rb_str_new(beg, sc->p - beg);
 }
 
-#define ATOM_SPECIAL  "()<>[]@,;:\"\\."
-#define TOKEN_SPECIAL "()<>[]@,;:\"\\/?="
-
-static VALUE
-scan_atom(sc)
-    struct mails *sc;
-{
-    return scan_word(sc, ATOM_SPECIAL);
-}
-
-static VALUE
-scan_token(sc)
-    struct mails *sc;
-{
-    return scan_word(sc, TOKEN_SPECIAL);
-}
-
 
 #define BUFSIZE 256
 
 static VALUE
-scan_qstr(sc, term)
-    struct mails *sc;
-    char term;
+scan_quoted_word(sc)
+    struct scanner *sc;
 {
-    char *p;
     char buf[BUFSIZE];
-    VALUE ret = rb_str_new("", 0);
+    char *p;
+    char *save;
+    VALUE result = rb_str_new("", 0);
 
+    sc->p++;   /* discard first dquote */
     p = buf;
     while (sc->p < sc->pend) {
-        if (*sc->p == term) {
-            sc->p++;
-            rb_str_cat(ret, buf, p - buf);
-            return ret;
+        if (*sc->p == '"') {
+            sc->p++;   /* discard terminal dquote */
+            rb_str_cat(result, buf, p - buf);
+            return result;
         }
-        else if (*sc->p == '\\') {
-            sc->p++;
+        if (ISO2022_MODE_P(sc) && *sc->p == ESC) {
+            save = sc->p;
+            skip_iso2022jp_string(sc);
+            while (save < sc->p) {
+                *p++ = *save++;
+                if (p >= buf + BUFSIZE) {
+                    /* flush buffer */
+                    rb_str_cat(result, buf, BUFSIZE);
+                    p = buf;
+                }
+            }
+            continue;
         }
 
-        *p = *sc->p; p++; sc->p++;
+        if (*sc->p == '\\')
+            sc->p++;   /* discard quoting backslash */
+        *p++ = *sc->p++;
         if (p >= buf + BUFSIZE) {
-            rb_str_cat(ret, buf, BUFSIZE);
+            /* flush buffer */
+            rb_str_cat(result, buf, BUFSIZE);
             p = buf;
         }
     }
 
-    rb_raise(ScanError, "unterminated quoted/domlit");
+    rb_raise(ScanError, "unterminated quoted-word");
     return Qnil;
 }
 
 static VALUE
-scan_quoted(sc)
-    struct mails *sc;
+scan_domain_literal(sc)
+    struct scanner *sc;
 {
-    sc->p++;
-    return scan_qstr(sc, '"');
-}
+    char buf[BUFSIZE];
+    char *p;
+    VALUE result = rb_str_new("", 0);
 
-static VALUE
-scan_domlit(sc)
-    struct mails *sc;
-{
-    sc->p++;
-    return scan_qstr(sc, ']');
-}
+    p = buf;
+    while (sc->p < sc->pend) {
+        if (*sc->p == ']') {
+            *p++ = *sc->p++;
+            rb_str_cat(result, buf, p - buf);
+            return result;
+        }
 
-
-#if 0
-#define BOUNDARY_CHAR "()+_,-./\': ?="
-
-static VALUE
-scan_boundary(sc)
-    struct mails *sc;
-{
-    unsigned char *beg;
-
-    for (beg = sc->p; sc->p < sc->pend; sc->p++) {
-        unsigned char ch = (unsigned char)(*sc->p);
-
-        if (! (strchr(BOUNDARY_CHAR, ch) ||
-               (*sc->p >= 'A' && *sc->p <= 'Z') ||
-               (*sc->p >= 'a' && *sc->p <= 'z') ||
-               (*sc->p >= '0' && *sc->p <= '9')))
-            break;
+        if (*sc->p == '\\')
+            sc->p++;   /* discard backslash */
+        *p++ = *sc->p++;
+        if (p >= buf + BUFSIZE) {
+            /* flush buffer */
+            rb_str_cat(result, buf, BUFSIZE);
+            p = buf;
         }
     }
 
-    return rb_str_new(beg, sc->p - beg);
-}
-#endif
-
-
-#if 0
-static void
-fwd_eol(sc)
-    struct mails *sc;
-{
-    if (*sc->p == '\n') {
-        sc->p++;
-    }
-    else if (*sc->p == '\r') {
-        sc->p++;
-        if (*sc->p == '\n')
-            sc->p++;
-    }
-
-    return;
+    rb_raise(ScanError, "unterminated domain literal");
+    return Qnil;
 }
 
 
-static void
-bufadd(buf, beg, len, ret, p_var)
-    char *buf, *beg;
-    long len;
-    VALUE ret;
-    char **p_var;
-{
-    if (*p_var - buf + len > BUFSIZE) {
-        /* flush buffer */
-        rb_str_cat(ret, buf, *p_var - buf);
-        *p_var = buf;
-    }
-    if (len > BUFSIZE) {
-        rb_str_cat(ret, beg, len);
-        *p_var = buf;
-    }
-    else {
-        memcpy(*p_var, beg, len);
-        *p_var += len;
-    }
-}
-#endif
-
 static VALUE
 scan_comment(sc)
-    struct mails *sc;
+    struct scanner *sc;
 {
     int nest = 1;
     char *p;
@@ -335,11 +287,11 @@ scan_comment(sc)
     sc->p++;
     p = sc->p;
     while (sc->p < sc->pend) {
-        if (*sc->p == ESC) {
-            fwd_jis(sc);
+        if (ISO2022_MODE_P(sc) && *sc->p == ESC) {
+            skip_iso2022jp_string(sc);
         }
         else if (IS_JCHAR(*sc->p)) {
-            fwd_jstr(sc);
+            skip_japanese_string(sc);
         }
         else {
             switch (*sc->p) {
@@ -373,22 +325,9 @@ scan_comment(sc)
 }
 
 
-#define IS_LWSP(ch) (ch == ' ' || ch == '\t' || ch == '\n' || ch == '\r')
-#define IS_DIGIT(ch) (ch >= '0' && ch <= '9')
-#define IS_ATOMCHAR(ch) \
-    ((ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z') || \
-     (ch >= '0' && ch <= '9') || (ch == ESC) || IS_JCHAR(ch) || \
-     strchr("#!$%&'`*+-{|}~^/=?", ch))
-
-#define IS_TOKENCHAR(ch) \
-    ((ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z') || \
-     (ch >= '0' && ch <= '9') || (ch == ESC) || IS_JCHAR(ch) || \
-     strchr("#!$%&'`*+-{|}~^.", ch))
-
-
 static void
 skip_lwsp(sc)
-    struct mails *sc;
+    struct scanner *sc;
 {
     while (sc->p < sc->pend) {
         if (IS_LWSP(*sc->p)) sc->p++;
@@ -396,21 +335,16 @@ skip_lwsp(sc)
     }
 }
 
-#define IS_ALPHA(ch) ((ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch >= 'Z'))
-#define IS_UPPER(ch) (ch >= 'A' && ch >= 'Z')
-#define TO_LOWER(ch) (IS_UPPER(ch) ? ch + 32 : ch)
-
 static int
 nccmp(a, b)
     char *a, *b;
 {
     while (*a && *b) {
-        if (*a != *b)
-            if (TO_LOWER(*a) != TO_LOWER(*b))
-                return 0;
+        if ((*a != *b) && (TO_LOWER(*a) != TO_LOWER(*b)))
+            return 0;
         a++; b++;
     }
-    return (*a == *b) ? 1 : 0;
+    return (*a == *b);
 }
 
 static int
@@ -433,7 +367,7 @@ static VALUE tok_from, tok_by, tok_via, tok_with, tok_id, tok_for;
 
 static VALUE
 atomsym(sc, str)
-    struct mails *sc;
+    struct scanner *sc;
     VALUE str;
 {
     if (digit_p(str)) {
@@ -452,8 +386,8 @@ atomsym(sc, str)
 }
 
 static void
-debugout(sc, sym, val)
-    struct mails *sc;
+debug_print(sc, sym, val)
+    struct scanner *sc;
     VALUE sym, val;
 {
     VALUE s;
@@ -468,14 +402,14 @@ static VALUE
 mails_scan(self)
     VALUE self;
 {
-    struct mails *sc;
+    struct scanner *sc;
     VALUE arr;
     int debug;
 
 #define PASS(s,v) \
-    if (debug) debugout(sc,s,v);\
-    RARRAY(arr)->ptr[0] = s;\
-    RARRAY(arr)->ptr[1] = v;\
+    if (debug) debug_print(sc,s,v);\
+    rb_ary_store(arr, 0, s);\
+    rb_ary_store(arr, 1, v);\
     rb_yield(arr)
 
 #define D(m) if (debug) puts(m)
@@ -485,7 +419,7 @@ mails_scan(self)
         rb_raise(ScanError, "Mails#scan called before reset");
     }
     arr = rb_assoc_new(Qnil, Qnil);
-    debug = sc->flags & MAILS_DEBUG;
+    debug = sc->flags & MODE_DEBUG;
 
     while (sc->p < sc->pend) {
         D("new loop");
@@ -497,14 +431,18 @@ mails_scan(self)
         }
 
         if (MIME_MODE_P(sc)) {
-            if (IS_TOKENCHAR(*sc->p)) {
+            if (IS_TOKENCHAR(*sc->p) ||
+                    (ISO2022_MODE_P(sc) && (*sc->p == ESC)) ||
+                    IS_JCHAR(*sc->p)) {
                 D("token");
                 PASS(tok_token, scan_token(sc));
                 continue;
             }
         }
         else {
-            if (IS_ATOMCHAR(*sc->p)) {
+            if (IS_ATOMCHAR(*sc->p) ||
+                    (ISO2022_MODE_P(sc) && (*sc->p == ESC)) ||
+                    IS_JCHAR(*sc->p)) {
                 VALUE tmp;
                 D("atom");
                 tmp = scan_atom(sc);
@@ -515,7 +453,7 @@ mails_scan(self)
 
         if (*sc->p == '"') {
             if (debug) puts("quoted");
-            PASS(tok_quoted, scan_quoted(sc));
+            PASS(tok_quoted, scan_quoted_word(sc));
         }
         else if (*sc->p == '(') {
             VALUE c;
@@ -526,7 +464,7 @@ mails_scan(self)
         }
         else if (*sc->p == '[') {
             D("domlit");
-            PASS(tok_domlit, scan_domlit(sc));
+            PASS(tok_domlit, scan_domain_literal(sc));
         }
         else {
             VALUE ch;
@@ -542,8 +480,14 @@ mails_scan(self)
 }
 
 
+/*
+------------------------------------------------------------------
+                         ruby interface
+------------------------------------------------------------------
+*/
+
 static VALUE
-intn(str)
+cstr2symbol(str)
     char *str;
 {
     ID tmp;
@@ -561,11 +505,9 @@ Init_cmails()
 {
     VALUE TMail;
     VALUE tmp;
-    ID sym_tmail = rb_intern("TMail");
-    ID sym_scerr = rb_intern("ScanError");
 
-    if (rb_const_defined(rb_cObject, sym_tmail))
-        TMail = rb_const_get(rb_cObject, sym_tmail);
+    if (rb_const_defined(rb_cObject, rb_intern("TMail")))
+        TMail = rb_const_get(rb_cObject, rb_intern("TMail"));
     else
         TMail = rb_define_module("TMail");
     MailScanner = rb_define_class_under(TMail, "Scanner_C", rb_cObject);
@@ -579,21 +521,21 @@ Init_cmails()
     rb_define_method(MailScanner, "debug", mails_debug_get, 0);
     rb_define_method(MailScanner, "debug=", mails_debug_set, 1);
 
-    if (rb_const_defined(rb_cObject, sym_scerr))
-        ScanError = rb_const_get(rb_cObject, sym_scerr);
+    if (rb_const_defined(TMail, rb_intern("SyntaxError")))
+        ScanError = rb_const_get(rb_cObject, rb_intern("SyntaxError"));
     else
-        ScanError = rb_define_class("ScanError", rb_eStandardError);
+        ScanError = rb_define_class_under(TMail, "SyntaxError", rb_eStandardError);
 
-    tok_atom   = intn("ATOM");
-    tok_digit  = intn("DIGIT");
-    tok_token  = intn("TOKEN");
-    tok_quoted = intn("QUOTED");
-    tok_domlit = intn("DOMLIT");
+    tok_atom   = cstr2symbol("ATOM");
+    tok_digit  = cstr2symbol("DIGIT");
+    tok_token  = cstr2symbol("TOKEN");
+    tok_quoted = cstr2symbol("QUOTED");
+    tok_domlit = cstr2symbol("DOMLIT");
 
-    tok_from = intn("FROM");
-    tok_by   = intn("BY");
-    tok_via  = intn("VIA");
-    tok_with = intn("WITH");
-    tok_id   = intn("ID");
-    tok_for  = intn("FOR");
+    tok_from   = cstr2symbol("FROM");
+    tok_by     = cstr2symbol("BY");
+    tok_via    = cstr2symbol("VIA");
+    tok_with   = cstr2symbol("WITH");
+    tok_id     = cstr2symbol("ID");
+    tok_for    = cstr2symbol("FOR");
 }
